from typing import List

import dask
import numpy
import numpy as np
from arbol.arbol import aprint, asection
from dask.distributed import Client
from dask_cuda import LocalCUDACluster

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.multiview_lightsheet.fusion.simview import SimViewFusion
from dexp.processing.registration.model.model_io import model_list_to_file
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel
from dexp.utils.slicing import slice_from_shape


def dataset_register(dataset,
                     model_path,
                     channels,
                     slicing,
                     microscope,
                     equalise,
                     zero_level,
                     clip_too_high,
                     fusion,
                     fusion_bias_strength_i,
                     dehaze_size,
                     registration_edge_filter,
                     max_proj,
                     workers,
                     workers_backend,
                     devices,
                     stop_at_exception=True):

    views = {
        channel.split('-')[-1]: dataset.get_array(channel, per_z_slice=False)
        for channel in channels
    }

    with asection(f"Views:"):
        for channel, view in views.items():
            aprint(f"View: {channel} of shape: {view.shape} and dtype: {view.dtype}")

    key = list(views.keys())[0]
    print(f"Slicing with: {slicing}")
    _, volume_slicing, time_points = slice_from_shape(views[key].shape, slicing)

    if microscope == 'simview':
        views = SimViewFusion.validate_views(views)
 
    @dask.delayed
    def process(i):
        tp = time_points[i]
        try:
            with asection(f"Loading channels {channel} for time point {i}/{len(time_points)}"):
                views_tp = {k: np.asarray(view[tp][volume_slicing]) for k, view in views.items()}

            with BestBackend(exclusive=True, enable_unified_memory=True):
                if microscope == 'simview':
                    fuse_obj = SimViewFusion(registration_model=None,
                                             equalise=equalise,
                                             equalisation_ratios=[None, None, None],
                                             zero_level=zero_level,
                                             clip_too_high=clip_too_high,
                                             fusion=fusion,
                                             fusion_bias_exponent=2,
                                             fusion_bias_strength_i=fusion_bias_strength_i,
                                             fusion_bias_strength_d=0.0,
                                             dehaze_before_fusion=True,
                                             dehaze_size=dehaze_size,
                                             dehaze_correct_max_level=True,
                                             dark_denoise_threshold=0,
                                             dark_denoise_size=0,
                                             butterworth_filter_cutoff=0.0,
                                             flip_camera1=True)

                    C0Lx, C1Lx = fuse_obj.preprocess(**views_tp)
                    del views_tp
                    Backend.current().clear_memory_pool()
                    fuse_obj.compute_registration(C0Lx, C1Lx, mode='projection' if max_proj else 'full',
                                                  edge_filter=registration_edge_filter,
                                                  crop_factor_along_z=0.3)
                    del C0Lx, C1Lx
                    Backend.current().clear_memory_pool()
                    model = fuse_obj.registration_model.to_numpy()
                else:
                    raise NotImplementedError

            aprint(f"Done processing time point: {i}/{len(time_points)} .")

        except Exception as error:
            aprint(error)
            aprint(f"Error occurred while processing time point {i} !")
            import traceback
            traceback.print_exc()

            if stop_at_exception:
                raise error

        return model

    cluster = LocalCUDACluster(CUDA_VISIBLE_DEVICES=devices)
    client = Client(cluster)
    aprint('Dask Client', client)

    lazy_computations = []
    for i in range(len(time_points)):
        lazy_computations.append(process(i))

    models = dask.compute(*lazy_computations)

    mode_model = compute_median_translation(models)
    model_list_to_file(model_path, [mode_model])
    client.close()


def compute_median_translation(models: List[TranslationRegistrationModel]) \
        -> TranslationRegistrationModel:
    """
    Computes the median of each axis over all registration models and
    assigns the confidence to the one with smallest distance to it.
    """
    models_shift = np.vstack([m.shift_vector for m in models if not numpy.isnan(m.confidence)])
    translation_mode = np.median(models_shift, axis=0)
    mode_model = TranslationRegistrationModel(translation_mode)
    distance = [m.change_relative_to(mode_model) for m in models]
    mode_model.confidence = models[np.argmin(distance)].confidence
    aprint(f'{len(models) - len(models_shift)} models with NaN confidence found.')
    aprint(f"Median model confidence of {mode_model.confidence}")
    return mode_model




from typing import Callable, Dict

import numpy
from numpy.typing import ArrayLike
from dexp.utils import xpArray
from dexp.processing.backends.backend import Backend

import itertools


def j_invariant_grid_search(image: xpArray,
                            function: Callable,
                            loss_fun: Callable[[xpArray, xpArray], float],
                            grid = Dict[str, ArrayLike],
                            proportion_mask: float = 0.01,
                            median_window: int = 3,
                            display: bool = False,
                        ) -> Dict[str, float]:
    
    backend = Backend.current()
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image = backend.to_backend(image)

    median = sp.ndimage.median_filter(image, size=median_window)
    mask = xp.random.uniform(size=image.shape) < proportion_mask

    # overwrite the values of the mask with median
    masked = image.copy()
    masked[mask] = median[mask]

    if display:
        import napari
        viewer = napari.view_image(backend.to_numpy(image), name='original')
        viewer.add_image(backend.to_numpy(median), name='median')
        viewer.add_image(backend.to_numpy(masked), name='mask')
        viewer.add_image(backend.to_numpy(mask), name='masked')

    history = []
    # compute grid search over possible values
    for params in itertools.product(*grid.values()):
        params = {k: p for k, p in zip(grid.keys(), params)}
        # predict results
        estimation = function(masked, **params)
        # compute loss function
        loss = loss_fun(estimation[mask], image[mask])
        loss = numpy.nan_to_num(loss, nan=1e12)

        name = f'{params}: loss={loss:0.3f}'
        if display:
            viewer.add_image(backend.to_numpy(estimation), name=name)
        print(name)
        history.append((loss, params))
    
    if display:
        napari.run()
    
    # return optimal (minimum)
    return min(history, key=lambda x: x[0])[1]




from typing import Callable, Dict, Optional
from numpy.typing import ArrayLike

from pyotf.phaseretrieval import PhaseRetrievalResult, retrieve_phase
from pyotf.utils import prep_data_for_PR

from dexp.utils import xpArray
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.backends.backend import Backend

import numpy as np
from toolz import curry
from scipy.optimize import minimize


@curry
def _objfun(coefs: ArrayLike, observed: xpArray, deconved: xpArray,
            psf_fun: Callable) -> float:
        # TODO: deconved could be pre computed in the freq domain
        
        backend = Backend.current()

        psf = psf_fun(coefs)
        psf = backend.to_backend(psf)

        blurred = fft_convolve(deconved, psf, mode='reflect')
        blurred -= blurred.min()
        blurred /= blurred.max()

        return ((observed - blurred) ** 2).mean().item()


def blind_deconvolution(image: xpArray,
                        observed_psf: xpArray,
                        deconv_fun: Callable[[xpArray, xpArray], xpArray],
                        microscope_params: Dict,
                        n_iterations: int,
                        n_zernikes: int = 15,
                        display: bool = False,
                    ) -> xpArray:
    
    if display:
        import napari
        viewer = napari.Viewer()

    backend = Backend.current()

    image = backend.to_backend(image)

    psf = prep_data_for_PR(observed_psf)
    phase_retriv: PhaseRetrievalResult = retrieve_phase(psf, params=microscope_params)
    phase_retriv.fit_to_zernikes(n_zernikes)

    def get_psf(coefs: Optional[ArrayLike] = None) -> ArrayLike:
        if coefs is not None:
            phase_retriv.zd_result.pcoefs = coefs[:len(coefs) // 2]
            phase_retriv.zd_result.mcoefs = coefs[len(coefs) // 2:]
        psf = phase_retriv.generate_zd_psf()
        return (psf / psf.sum()).astype(image.dtype)

    if display:
        viewer.add_image(backend.to_numpy(image), name='Input Image')
        viewer.add_image(observed_psf, name='Observed PSF')
        viewer.add_image(psf, name='Prepared PSF')
 
    psf = get_psf()
    if display:
       viewer.add_image(psf, name='Estimated PSF')

    coefs = np.concatenate((phase_retriv.zd_result.pcoefs, phase_retriv.zd_result.mcoefs))

    for i in range(n_iterations):

        deconv = deconv_fun(image, psf)

        objfun = _objfun(observed=image, deconved=deconv, psf_fun=get_psf)
        opt_res = minimize(objfun, x0=coefs, method='L-BFGS-B')

        old_coefs = coefs
        coefs = opt_res.x
        psf = get_psf(coefs)

        # FIXME
        print(np.square(old_coefs - coefs).mean())

        if display:
            viewer.add_image(backend.to_numpy(deconv), name=f'Deconv i={i}')
            viewer.add_image(psf, name=f'Est. PSF i={i}')
    
    if display:
        napari.run()

    return deconv



from typing import Sequence, Tuple, List, Optional

import dask
import numpy
import scipy
import functools
from pathlib import Path

from arbol.arbol import aprint
from arbol.arbol import asection
from dask.distributed import Client
from dask_cuda import LocalCUDACluster

from dexp.datasets.base_dataset import BaseDataset
from dexp.optics.psf.standard_psfs import nikon16x08na, olympus20x10na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution
from dexp.processing.deconvolution.admm_deconvolution import admm_deconvolution
from dexp.processing.utils.scatter_gather_i2i import scatter_gather_i2i
from dexp.utils.slicing import slice_from_shape


def dataset_deconv(dataset: BaseDataset,
                   dest_path: str,
                   channels: Sequence[str],
                   slicing,
                   store: str = 'dir',
                   compression: str = 'zstd',
                   compression_level: int = 3,
                   overwrite: bool = False,
                   tilesize: Optional[Tuple[int]] = None,
                   method: str = 'lr',
                   num_iterations: int = 16,
                   max_correction: int = 16,
                   power: float = 1,
                   blind_spot: int = 0,
                   back_projection: Optional[str] = None,
                   wb_order: int = 5,
                   psf_objective: str = 'nikon16x08na',
                   psf_na: float = 0.8,
                   psf_dxy: float = 0.485,
                   psf_dz: float = 2,
                   psf_xy_size: int = 17,
                   psf_z_size: int = 17,
                   psf_show: bool = False,
                   scaling: Optional[Tuple[float]] = None,
                   workers: int = 1,
                   workersbackend: str = '',
                   devices: Optional[List[int]] = None,
                   check: bool = True,
                   stop_at_exception: bool = True,
                   ):

    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(dest_path, mode, store, parent=dataset)

    # Default tile size:
    if tilesize is None:
        tilesize = 320 # very conservative

    # Scaling default value:
    if scaling is None:
        scaling = (1, 1, 1)
    sz, sy, sx = scaling
    aprint(f"Input images will be scaled by: (sz,sy,sx)={scaling}")

    # CUDA DASK cluster
    cluster = LocalCUDACluster(CUDA_VISIBLE_DEVICES=devices)
    client = Client(cluster)
    aprint('Dask Client', client)

    lazy_computation = []

    for channel in dataset._selected_channels(channels):
        array = dataset.get_array(channel)

        aprint(f"Slicing with: {slicing}")
        out_shape, volume_slicing, time_points = slice_from_shape(array.shape, slicing)

        out_shape = tuple(int(round(u*v)) for u, v in zip(out_shape, (1,)+scaling))
        dtype = numpy.float16 if method == 'admm' else array.dtype 

        # Adds destination array channel to dataset
        dest_array = dest_dataset.add_channel(name=channel,
                                              shape=out_shape,
                                              dtype=dtype,
                                              codec=compression,
                                              clevel=compression_level)

        #This is not ideal but difficult to avoid right now:
        sxy = (sx+sy)/2

        # PSF paraneters:
        psf_kwargs = {'dxy': psf_dxy / sxy,
                      'dz': psf_dz / sz,
                      'xy_size': int(round(psf_xy_size * sxy)),
                      'z_size': int(round(psf_z_size * sz))}

        aprint(f"psf_kwargs: {psf_kwargs}")

        # NA override:
        if psf_na is not None:
            aprint(f"Numerical aperture overridden to a value of: {psf_na}")
            psf_kwargs['NA'] = psf_na

        # choose psf from detection optics:
        if psf_objective == 'nikon16x08na':
            psf_kernel = nikon16x08na(**psf_kwargs)
        elif psf_objective == 'olympus20x10na':
            psf_kernel = olympus20x10na(**psf_kwargs)
        elif Path(psf_objective).exists():
            psf_kernel = numpy.load(psf_objective)
            if sz != 1.0 or sy != 1.0 or sx != 1.0:
                psf_kernel = scipy.ndimage.interpolation.zoom(psf_kernel, zoom=(sz, sy, sx), order=1)
            psf_z_size = psf_kernel.shape[0] + 10
            psf_xy_size = max(psf_kernel.shape[1:]) + 10
        else:
            raise RuntimeError(f'Object/path {psf_objective} not found.')

        # usefull for debugging:
        if psf_show:
            import napari
            viewer = napari.Viewer(title=f"DEXP | viewing PSF with napari", ndisplay=3)
            viewer.add_image(psf_kernel)
            napari.run()

        margins = max(psf_xy_size, psf_z_size)

        if method == 'lr':
            normalize = False
            convolve = functools.partial(fft_convolve,
                                         in_place=False,
                                         mode='reflect',
                                         internal_dtype=numpy.float32)

            def deconv(image):
                min_value = image.min()
                max_value = image.max()

                return lucy_richardson_deconvolution(image=image,
                                                     psf=psf_kernel,
                                                     num_iterations=num_iterations,
                                                     max_correction=max_correction,
                                                     normalise_minmax=(min_value, max_value),
                                                     power=power,
                                                     blind_spot=blind_spot,
                                                     blind_spot_mode='median+uniform',
                                                     blind_spot_axis_exclusion=(0,),
                                                     wb_order=wb_order,
                                                     back_projection=back_projection,
                                                     convolve_method=convolve
                                                     )
        elif method == 'admm':
            normalize = True
            def deconv(image):
                out = admm_deconvolution(
                    image,
                    psf=psf_kernel,
                    iterations=num_iterations,
                    derivative=2,
                )
                return out

        else:
            raise ValueError(f"Unknown deconvolution mode: {method}")


        @dask.delayed
        def process(i):
            tp = time_points[i]
            try:
                with asection(f'Deconvolving time point for time point {i}/{len(time_points)}'):
                    with asection(f"Loading channel: {channel}"):
                        tp_array = numpy.asarray(array[tp][volume_slicing])

                    with BestBackend(exclusive=True, enable_unified_memory=True):

                        if sz != 1.0 or sy != 1.0 or sx != 1.0:
                            with asection(f"Applying scaling {(sz, sy, sx)} to image."):
                                sp = Backend.get_sp_module()
                                tp_array = Backend.to_backend(tp_array)
                                tp_array = sp.ndimage.interpolation.zoom(tp_array, zoom=(sz, sy, sx), order=1)
                                tp_array = Backend.to_numpy(tp_array)

                        with asection(f"Deconvolving image of shape: {tp_array.shape}, with tile size: {tilesize}, margins: {margins} "):
                            aprint(f"Number of iterations: {num_iterations}, back_projection:{back_projection}, ")
                            tp_array = scatter_gather_i2i(deconv, tp_array, tiles=tilesize, margins=margins,
                                                          normalise=normalize, internal_dtype=dtype)

                        with asection(f"Moving array from backend to numpy."):
                            tp_array = Backend.to_numpy(tp_array, dtype=dest_array.dtype, force_copy=False)

                    with asection(f"Saving deconvolved stack for time point {i}, shape:{tp_array.shape}, dtype:{array.dtype}"):
                        dest_dataset.write_stack(channel=channel,
                                                 time_point=i,
                                                 stack_array=tp_array)

                    aprint(f"Done processing time point: {i}/{len(time_points)} .")

            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {i} !")
                import traceback
                traceback.print_exc()

                if stop_at_exception:
                    raise error

        for i in range(len(time_points)):
            lazy_computation.append(process(i))

    dask.compute(*lazy_computation)

    # Dataset info:
    aprint(dest_dataset.info())

    # Check dataset integrity:
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()
    client.close()



from typing import Sequence, Union, List

import dask
from arbol import aprint
from dask.array import reshape

from dexp.datasets.base_dataset import BaseDataset


def dataset_view(input_dataset: BaseDataset,
                 channels: Sequence[str],
                 slicing,
                 aspect: float,
                 contrast_limits: Union[List[int], List[float]],
                 colormap: str,
                 name: str,
                 windowsize: int,
                 projections_only: bool,
                 volume_only: bool,
                 rescale_time: bool):

    import napari
    from napari.layers.utils._link_layers import link_layers, _get_common_evented_attributes

    if rescale_time:
        nb_time_points = [input_dataset.shape(channel)[0] for channel in input_dataset.channels()]
        max_time_points = max(nb_time_points)
        time_scales = [round(max_time_points / n) for n in nb_time_points]
    else:
        time_scales = [1] * len(input_dataset.channels())

    viewer = napari.Viewer(title=f"DEXP | viewing with napari: {name} ", ndisplay=2)
    viewer.grid.enabled = True
    viewer.window.resize(windowsize + 256, windowsize)

    for time_scale, channel in zip(time_scales, channels):
        aprint(f"Channel '{channel}' shape: {input_dataset.shape(channel)}")
        aprint(input_dataset.info(channel))

        array = input_dataset.get_array(channel, wrap_with_dask=True)

        layers = []
        try:
            if not volume_only:
                for axis in range(array.ndim - 1):
                    proj_array = input_dataset.get_projection_array(channel, axis=axis, wrap_with_dask=True)

                    # if the data format does not support projections we skip:
                    if proj_array is None:
                        continue

                    if axis == 1:
                        proj_array = dask.array.flip(proj_array, 1)  # flipping y
                    elif axis == 2:
                        proj_array = dask.array.rot90(proj_array, axes=(2, 1))  #

                    shape = (proj_array.shape[0], 1,) + proj_array.shape[1:]
                    proj_array = reshape(proj_array, shape=shape)

                    if proj_array is not None:
                        proj_layer = viewer.add_image(proj_array,
                                                        name=channel + '_proj_' + str(axis),
                                                        contrast_limits=contrast_limits,
                                                        blending='additive',
                                                        colormap=colormap, )
                        layers.append(proj_layer)

                        if aspect is not None:
                            if axis == 0:
                                proj_layer.scale = (time_scale, 1, 1, 1)
                            elif axis == 1:
                                proj_layer.scale = (time_scale, 1, aspect, 1)
                            elif axis == 2:
                                proj_layer.scale = (time_scale, 1, 1, aspect)  # reordered due to rotation

                            aprint(f"Setting aspect ratio for projection (layer.scale={proj_layer.scale})")

        except KeyError:
            aprint("Warning: can't find projections!")

        if not projections_only:

            if slicing:
                array = array[slicing]

            aprint(f"Adding array of shape={array.shape} and dtype={array.dtype} for channel '{channel}'.")

            # flip x for second camera:
            if 'C1' in channel:
                array = dask.array.flip(array, -1)

            layer = viewer.add_image(array,
                                        name=channel,
                                        contrast_limits=contrast_limits,
                                        blending='additive',
                                        colormap=colormap,
                                        attenuation=0.04,
                                        rendering='attenuated_mip')
            layers.append(layer)

            if aspect is not None:
                if array.ndim == 3:
                    layer.scale = (aspect, 1, 1)
                elif array.ndim == 4:
                    layer.scale = (time_scale, aspect, 1, 1)
                    aprint(f'Setting time scale to {time_scale}')
                aprint(f"Setting aspect ratio to {aspect} (layer.scale={layer.scale})")

        if layers:
            attr = _get_common_evented_attributes(layers)
            attr.remove('visible')
            link_layers(layers, attr)

    napari.run()



import os
from os.path import join
from typing import Sequence, Union

from arbol.arbol import aprint, asection
from joblib import Parallel, delayed
from tifffile import memmap

from dexp.datasets.base_dataset import BaseDataset
from dexp.io.io import tiff_save


def dataset_tiff(dataset: BaseDataset,
                 dest_path: str,
                 channels: Sequence[str],
                 slicing,
                 overwrite: bool = False,
                 project: Union[int, bool] = False,
                 one_file_per_first_dim: bool = False,
                 clevel: int = 0,
                 workers: int = 1,
                 workersbackend: str = '',
                 stop_at_exception: bool = True):


    selected_channels = dataset._selected_channels(channels)

    aprint(f"getting Dask arrays for channels {selected_channels}")
    arrays = list([dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True) for channel in selected_channels])

    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        arrays = list([array[slicing] for array in arrays])
        aprint(f"Done slicing.")

    if workers == -1:
        workers = max(1, os.cpu_count() // abs(workers))
    aprint(f"Number of workers: {workers}")

    if one_file_per_first_dim:
        aprint(f"Saving one TIFF file for each tp (or Z if already sliced) to: {dest_path}.")

        os.makedirs(dest_path, exist_ok=True)

        def process(tp):
            try:
                with asection(f'Saving time point {tp}: '):
                    for channel, array in zip(selected_channels, arrays):
                        tiff_file_path = join(dest_path, f"file{tp}_{channel}.tiff")
                        if overwrite or not os.path.exists(tiff_file_path):
                            stack = array[tp].compute()

                            if project is not False and type(project) == int:
                                # project is the axis for projection, but here we are not considering the T dimension anymore...
                                aprint(f"Projecting along axis {project}")
                                stack = stack.max(axis=project)

                            aprint(f"Writing time point: {tp} of shape: {stack.shape}, dtype:{stack.dtype} as TIFF file: '{tiff_file_path}', with compression: {clevel}")
                            tiff_save(tiff_file_path, stack, compress=clevel)
                            aprint(f"Done writing time point: {tp} !")
                        else:
                            aprint(f"File for time point (or z slice): {tp} already exists.")
            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {tp} !")
                import traceback
                traceback.print_exc()

                if stop_at_exception:
                    raise error

        if workers > 1:
            Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp) for tp in range(0, arrays[0].shape[0]))
        else:
            for tp in range(0, arrays[0].shape[0]):
                process(tp)

    else:

        for channel, array in zip(selected_channels, arrays):
            if len(selected_channels) > 1:
                tiff_file_path = f"{dest_path}_{channel}.tiff"
            else:
                tiff_file_path = f"{dest_path}.tiff"

            if not overwrite and os.path.exists(tiff_file_path):
                aprint(f"File {tiff_file_path} already exists! Set option -w to overwrite.")
                return

            with asection(f"Saving array ({array.shape}, {array.dtype}) for channel {channel} into TIFF file at: {tiff_file_path}:"):

                shape = array.shape

                if project is not False and type(project) == int:
                    shape = list(shape)
                    shape.pop(1+project)
                    shape = tuple(shape)

                memmap_image = memmap(tiff_file_path, shape=shape, dtype=array.dtype, bigtiff=True, imagej=True)

                def process(tp):
                    aprint(f"Processing time point {tp}")
                    stack = array[tp].compute()

                    if project is not False and type(project) == int:
                        # project is the axis for projection, but here we are not considering the T dimension anymore...
                        aprint(f"Projecting along axis {project}")
                        stack = stack.max(axis=project)

                    memmap_image[tp] = stack

                if workers > 1:
                    Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp) for tp in range(0, array.shape[0]))
                else:
                    for tp in range(0, array.shape[0]):
                        process(tp)

                memmap_image.flush()
                del memmap_image

## NOTES: color coded max projection:
# > data = numpy.random.randint(0, 255, (256, 256, 3), 'uint8')
# >>> imwrite('temp.tif', data, photometric='color')
#
# https://colorcet.holoviz.org/user_guide/index.html
# https://github.com/MMesch/cmap_builder
#




from typing import Optional

import numpy
import scipy

from dexp.utils import xpArray
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend



def inversion_deconvolution(image: xpArray,
                            psf: xpArray,
                            epsilon: float = 1e-12,
                            mode: str = 'wrap',
                            internal_dtype: Optional[numpy.dtype] = None) -> xpArray:
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if image.ndim != psf.ndim:
        raise ValueError("The image and PSF must have same number of dimensions!")

    if internal_dtype is None:
        internal_dtype = numpy.float32

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype
    image = Backend.to_backend(image, dtype=internal_dtype)
    psf = Backend.to_backend(psf, dtype=internal_dtype)

    backproj = xp.flip(psf)

    if mode != 'wrap':
        pad_width = tuple((tuple((s // 2, s // 2)) for s in backproj.shape))
        image = xp.pad(image, pad_width=pad_width, mode=mode)

    s1 = numpy.array(image.shape)
    s2 = numpy.array(backproj.shape)

    shape = tuple(s1 + s2 - 1)
    fsize = tuple(scipy.fftpack.next_fast_len(x) for x in tuple(shape))

    image_fft = sp.fft.rfftn(image, fsize)
    backproj_fft = sp.fft.rfftn(backproj, fsize)

    # deconv_fft = (backproj_fft.conjugate() * image_fft + epsilon) / (backproj_fft * backproj_fft + epsilon)
    deconv_fft = (image_fft + epsilon) / (backproj_fft + epsilon)

    import napari
    viewer = napari.Viewer()
    viewer.add_image(xp.fft.fftshift(image_fft).real.get(), name='image')
    viewer.add_image(xp.fft.fftshift(backproj_fft).real.get(), name='backprojector')
    viewer.add_image(xp.fft.fftshift(deconv_fft).real.get(), name='deconvolved')

    napari.run()

    deconv = sp.fft.irfftn(deconv_fft)

    fslice = tuple([slice(0, int(sz)) for sz in shape])
    deconv = deconv[fslice]

    newshape = numpy.asarray(image.shape)
    currshape = numpy.array(deconv.shape)
    startind = (currshape - newshape) // 2
    endind = startind + newshape
    myslice = [slice(startind[k], endind[k]) for k in range(len(endind))]

    deconv = deconv[tuple(myslice)]

    if mode != 'wrap':
        slicing = tuple(slice(s // 2, -(s // 2)) for s in backproj.shape)
        deconv = deconv[slicing]

    return deconv.astype(dtype=original_dtype, copy=False)
